/** This file is part of GeodeMAPS, a program for computing the geodesic distance between phylogenetic trees.
    Copyright (C) 2008  Megan Owen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


package distanceAlg1;

// TODO:  create sameAs method and fix getCommonEdges
import java.util.*;

public class Bipartition {
	
	protected BitSet edge;

	public Bipartition(){
		edge = new BitSet();
	}
	
	public Bipartition(BitSet edge) {
		this.edge = edge;
	}
	
	public BitSet getEdge() {
		return edge;
	}

	public boolean isZero() {
/*		if (edge.isEmpty()) {
			System.out.println("in isZero, edge is " + this.edge + " and will return " + edge.isEmpty());
		}*/
		return edge.isEmpty();
	}
	
	public void setEdge(BitSet edge) {
		this.edge = edge;
	}

	/** Adds a 1 to the edge, or changes an element from one block to the other in the partition.
	 * Adds the 1 at coordinate one, from the right.
	 * Precondition:  this edge contains a 0 at position one
	 * 
	 * @param one
	 */
	public void addOne(int one) {
//		System.out.println("edge is " + edge);
		edge.set(one);
//		System.out.println("edge with flipped bit in " + one + " is " + edge);
	}

	/** Removes a 1 from the edge.  In other words, transfers the element one to the other (zero)
	 * block in the partition.
	 * Precondition:  this edge contains a 1 at position one 
	 * @param leafNum
	 */
	public void removeOne(int one) {
		edge.clear(one);
	}

	/** = e1 and e2 are disjoint edges
	 * Precondition:  the leaf2NumMap for the trees containing e1 and e2 are the same.
	 * @param e1
	 * @param e2
	 * @return
	 */
	public boolean disjointFrom(Bipartition e) {
//		return (this.edge.and(e.edge).compareTo(BigInteger.ZERO) == 0 );
//		System.out.println("" + this.edge + " and " + e.edge + " are disjoint: " + !this.edge.intersects(e.edge));
		return !this.edge.intersects(e.edge);
//		return (this.edge & e.edge) == 0;
	}

	/** = this edge contains e
	 * Precondition:  the leaf2NumMap for the tree containing this edge and the tree containing e are the same.
	 * XXX:  are we changing e outside of this method?
	 * @param e
	 * @return
	 */
	public boolean contains(Bipartition e) {
		
		BitSet edgeClone = (BitSet) e.edge.clone();
		edgeClone.and(this.edge);
//		System.out.println("edgeClone is " + edgeClone);
//		System.out.println("" + this.edge + " contains " + e.edge + "? " + edgeClone.equals(e.edge) );
		return edgeClone.equals(e.edge);
//		return (this.edge.and(e.edge).compareTo(e.edge) == 0);
//		return (edge & e.edge) == e.edge;
	}

	/** = this edge contains a 1 in the bit corresponding to the i-th column
	 * 
	 * @param i
	 * @return
	 */
	public boolean contains(int i) {
//		System.out.println("" + this.edge + " contains edge " + i + ": " + this.edge.get(i));
		return this.edge.get(i);
		// return this.edge.testBit(i);
//		return this.contains( new Bipartition((long) Math.pow(2,i)) );
	}
		
	/** = this edge properly contains e  (ie. contains e, but is not equal to e)
	 * 
	 * @param e
	 * @return
	 */
	public boolean properlyContains(Bipartition e) {
		return this.contains(e) && !this.equals(e);
	}
	
	/** = this edge crosses e
	 * 
	 * @param e
	 * @return
	 */
	public boolean crosses(Bipartition e) {
		return !(disjointFrom(e) || this.contains(e) || e.contains(this));
	}

	/** Returns true if e contains ones exactly where this edges doesn't.
	 * numLeaves gives the total number of ones each should contain.
	 * @param e
	 * @param numLeaves
	 * @return
	 */
/*	public boolean isComplementOf(Bipartition e, int numLeaves) {
//		System.out.println("Sum of edges " + e.edge + " and " + this.edge + " is " + (e.edge + this.edge) + " and want " + ((long)Math.pow(2,numLeaves) - 1));
		if ( (e.edge + this.edge) == ((long)Math.pow(2,numLeaves) - 1) )  {
			return true;
		}
		else {
			return false;
		}
	}*/
	
	/** Changes e to be its complement.
	 *  XXX:  I think this is a slow method; make faster
	 * @param e
	 * @param numLeaves
	 * @return
	 */
	public void complement (int numLeaves) {
//		this.edge = (long)Math.pow(2,numLeaves) - 1 - this.edge;
		for (int i = 0; i < numLeaves; i++) {
			this.edge.flip(i);
		}
	}
	
	/** Coverts edge into a 0-1 vector.  XXX: leading zeros not shown.
	 *  XXX:  maybe fix???
	 */
	public String toString() {
//		return Long.toBinaryString(edge);
//		return "" + this.edge.toByteArray();
		return this.edge.toString();
	}

	public static String toStringVerbose(BitSet edge, Vector<String> leaf2NumMap) {
/*		String toDisplay = "";
		
		String s = Long.toBinaryString(edge);
		// we now have a 0-1 string, but it is reversed.  ie. 100 should become {2} and 1100 should become {3,2}
		for (int i = s.length()-1; i >=0; i--) {
			if (s.charAt(i) == '1') {
				toDisplay = toDisplay + leaf2NumMap.get(s.length() -1 - i) + "*";
			}
		}
		// remove the last *
		return toDisplay.substring(0, toDisplay.length() - 1); */
		String toDisplay = "";
		for (int i = 0; i < edge.length(); i++) {
			if (edge.get(i)) {
				toDisplay = toDisplay + leaf2NumMap.get(i) + ",";
			}
		}
		// remove the last ,
		return toDisplay.substring(0, toDisplay.length()-1);
//		return toDisplay;
	}
	
	/**
	 *  XXX:  check that e not changed outside this method
	 * @param e
	 * @return
	 */
	
	public boolean equals(Bipartition e) {
//		System.out.println("edge is " + edge + " and e is " + e.edge + " and we will return " + edge.equals(e.edge));
		return edge.equals(e.edge);
//		return (edge.compareTo(e.edge) == 0);
		// return edge == e.edge;
	}
	
	public Bipartition clone() {
		return new Bipartition((BitSet) edge.clone());
//		return new Bipartition(new BigInteger(this.edge.toString()));
	}	
}
