/** This file is part of GeodeMAPS, a program for computing the geodesic distance between phylogenetic trees.
    Copyright (C) 2008  Megan Owen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


package distanceAlg1;

import java.util.*;

/* XXX: (general problems)
1) we need a method to convert trees without the same leaf2NumMap to having the same one, if possible.

*/

public class PhyloTree {
	private Vector<PhyloTreeEdge> edges;	 // each element of edges represents an interior edge of the tree.
	private Vector<String> leaf2NumMap;   // the position the leaf name occurs in this vector corresponds to its coordinate in an edge vector.
							// ie. if leaf A is in position 5, then, a 1 in an edge vector at position 5 means that edges contains leaf A
							// standardized to always be in the natural order for strings
	private double[] leafEdgeLengths;
	private String newick;  // the newick format of the tree, if available
	
	// constructor
	// assume these trees are already rooted according to convention; 
	// this shouldn't be a problem, because we root any tree read in from a string according to convention
	public PhyloTree(Vector<PhyloTreeEdge> edges, Vector<String> leaf2NumMap, double[] leafEdgeLengths) {
		this.edges = edges;
		this.leaf2NumMap = leaf2NumMap;
		this.leafEdgeLengths = leafEdgeLengths;
	}
	
	// constructor
	// assume these trees are already rooted according to convention; 
	// this shouldn't be a problem, because we root any tree read in from a string according to convention
	public PhyloTree(Vector<PhyloTreeEdge> edges, Vector<String> leaf2NumMap) {
		this.edges = edges;
		this.leaf2NumMap = leaf2NumMap;
	}
	
	/** Constructor - turns t into a PhyloTree, by storing the edge data in edges and creating a conversion chart between the leaf names
	 * and their position in edge vectors.  Basically each name is assigned a different number from 1 to # of leaves.
	 * Example of Newick Standard:  (B:6.0,(A:5.0,C:3.0,E:4.0):5.0,D:11.0);
	 * If tree is really unrooted, as passed in as argument and stored in TreeDistance class variable, 
	 * then reroot the tree so that the last leaf in leaf2NumMap is the root (and then everything automatically becomes one leaf smaller)
	 * Handles multifurcating trees.
	 * @param t representation of a tree in Newick standard.
	 */
	public PhyloTree(String t) {
		int leafNum = 0;
		double leafEdgeLength;
		//TODO:  do more tests?  
		// XXX: clean up this code!
		// TODO: put in exception handling to deal with malformed string gracefully
		// XXX: for now this method ignores edges lengths...
		LinkedList<PhyloTreeEdge> queue = new LinkedList<PhyloTreeEdge>();   // queue for keeping track of edges we are still adding leaves to
		edges = new Vector<PhyloTreeEdge>();
		leaf2NumMap = new Vector<String>();
		int endOfLeafEdgeLength = 0;
		
		newick = t;
		
		//pull off ';' if at end
		if (t.charAt(t.length()-1) == ';') {
			t = t.substring(0, t.length() -1);
		}
		
		// pull off the first and last brackets.
		t = t.substring(1,t.length()-1);
		
		int i= 0;
		// go through the string and pull out all the leaf labels:  any string between '(' and ':' or ',' and ':'
		while (i < t.length()) {
			// however, the first character might be the beginning of a leaf label
			if (i == 0 && t.charAt(i) != '(') {
				leaf2NumMap.add(t.substring(0, t.indexOf(":")));
				i = nextIndex(t, i, ",)");
			}
			else if ((t.charAt(i) == '(' || t.charAt(i) == ',')&&(t.charAt(i+1) != '(')) {
				leaf2NumMap.add(t.substring(i+1, t.substring(i).indexOf(":")+i));
				i = nextIndex(t, i, ",)");
			}
			else {
				i++;
			}
		}
		
		// sort the elements of leaf2NumMap
		Collections.sort(leaf2NumMap);  
		
		leafEdgeLengths = new double[leaf2NumMap.size()];
		
		// now go back, and associate leaves with edges.
		i = 0;
		while (i < t.length() && i > -1) {
			if (t.charAt(i) == '(') {
				queue.addFirst(new PhyloTreeEdge());
				if (t.charAt(i+1) != '(') {
					// get following leaf label, which ends with a :
					leafNum = leaf2NumMap.indexOf(t.substring(i+1, t.substring(i).indexOf(":")+i));
					
					// the following two lines get the leaf edge length for this leaf, and store it
		//			System.out.println("Substring is " + t.substring( t.substring(i+1).indexOf(':') + 1,Math.min(t.substring(i+1).indexOf(','), t.substring(i+1).indexOf(')')-1 ) ) );
					if ((t.substring(i+1).indexOf(',') > -1) && (t.substring(i+1).indexOf(')') > -1) ) { 
						if (t.substring(i+1).indexOf(',') <t.substring(i+1).indexOf(')') )  {
							endOfLeafEdgeLength = t.substring(i+1).indexOf(',') + i+1;
						}
						else {
							endOfLeafEdgeLength = t.substring(i+1).indexOf(')') + i+1;
						}
					}
					else if (t.substring(i+1).indexOf(')') > -1) {
						endOfLeafEdgeLength = t.substring(i+1).indexOf(')') + i+1;
					}
					else if (t.substring(i+1).indexOf(',') > -1) {
						endOfLeafEdgeLength = t.substring(i+1).indexOf(',') + i+1;
					}
					else {
						// we have removed the end bracket, so we are not finding that
						endOfLeafEdgeLength = t.length();
					}
					leafEdgeLength = Double.parseDouble(t.substring( (t.substring(i+1).indexOf(':') + i + 2),endOfLeafEdgeLength  )  );
					leafEdgeLengths[leafNum] = leafEdgeLength;  
					
					// we now want to add this leaf to all the edges in our queue.
					ListIterator<PhyloTreeEdge> iter = queue.listIterator();
					while (iter.hasNext()) {
						iter.next().addOne(leafNum);
					}

					// XXX for now, skip any length on the ending leaves
					i = nextIndex(t, i, ",)");
				}
				else {
					// we have at least two ( in a row
					i++;
				}
			}
				
			else if (t.charAt(i) == ')') {
				// Extract this edge's length, which is between the : following the ), and either a , or a ).
				queue.peek().setLength(Double.parseDouble(t.substring(i+2, nextIndex(t, i, ",)"))));
				edges.add(queue.poll());
				// increment i, so it is the next index after the length
				i = nextIndex(t, i, ",)");
			}
				
			else if (t.charAt(i) == ',') {
				// we are at the beginning of a leaf label OR an open bracket
				if (t.charAt(i+1) == '(') {
					i++;
				}
				else {
					leafNum = leaf2NumMap.indexOf(t.substring(i+1, t.substring(i).indexOf(":")+i));
					// get the leaf edge length
					if ((t.substring(i+1).indexOf(',') > -1) && (t.substring(i+1).indexOf(')') > -1) ) { 
						if (t.substring(i+1).indexOf(',') <t.substring(i+1).indexOf(')') )  {
							endOfLeafEdgeLength = t.substring(i+1).indexOf(',') + i+1;
						}
						else {
							endOfLeafEdgeLength = t.substring(i+1).indexOf(')') + i+1;
						}
					}
					else if (t.substring(i+1).indexOf(')') > -1) {
						endOfLeafEdgeLength = t.substring(i+1).indexOf(')') + i+1;
					}
					else if (t.substring(i+1).indexOf(',') > -1) {
						endOfLeafEdgeLength = t.substring(i+1).indexOf(',') + i+1;
					}
					else {
						// we have removed the end bracket, so we are not finding that
						endOfLeafEdgeLength = t.length();
					}
					
					leafEdgeLength = Double.parseDouble(t.substring( (t.substring(i+1).indexOf(':') + i + 2),endOfLeafEdgeLength  )  );
					leafEdgeLengths[leafNum] = leafEdgeLength;  
					
					
					// we now want to add this leaf to all the edges in our queue.
					ListIterator<PhyloTreeEdge> iter = queue.listIterator();
					while (iter.hasNext()) {
						iter.next().addOne(leafNum);
					}
					// XXX for now, skip any length on the ending leaves
					i = nextIndex(t, i, ",)");
				}
			}
			else {
				// we are at the start of a leaf label, that is a solitary leaf in the trees.  
				leafNum = leaf2NumMap.indexOf(t.substring(i+1, t.substring(i).indexOf(":")+i));
				// we now want to add this leaf to all the edges in our queue.
				ListIterator<PhyloTreeEdge> iter = queue.listIterator();
				while (iter.hasNext()) {
					iter.next().addOne(leafNum);
				}
				// XXX for now, skip any length on the ending leaves
				i = nextIndex(t, i, ",)");
			}
		}
		
		// if tree is really unrooted, reroot so that the last leaf in leaf2NumMap is the root
		if (!TreeDistance.rooted) {
			for (int k = 0; k < edges.size(); k++) {
				int lastLeaf = leaf2NumMap.size()-1;
				if (getEdge(k).contains(lastLeaf)) {
					getEdge(k).complement(lastLeaf +1);
				}
			}
		}
		
		for (int k = 0; k < edges.size(); k++) {
			edges.get(k).setOriginalEdge(edges.get(k).getEdge());
		}
/*		for (int k = 0; k < leaf2NumMap.size(); k++) {
			System.out.println("Leaf " + leaf2NumMap.get(k) + " has length " + leafEdgeLengths[k]);
		}
		System.out.println("leaf2NumMap is " + leaf2NumMap); */
	}
	
	// Getters and Setters
	public Vector<PhyloTreeEdge> getEdges() {
		return edges;
	}

	/** Returns the edge at position i in the edge vector
	 * 
	 * @param i
	 * @return
	 */
	public PhyloTreeEdge getEdge(int i) {
		return edges.get(i);
	}
	
	public void setEdges(Vector<PhyloTreeEdge> edges) {
		this.edges = edges;
	}


	public Vector<String> getLeaf2NumMap() {
		return leaf2NumMap;
	}

	public double getLengthOfEdge(Bipartition edge) {
		Iterator<PhyloTreeEdge> edgesIter = edges.iterator();
		while (edgesIter.hasNext()){
			PhyloTreeEdge e = (PhyloTreeEdge) edgesIter.next();
			if (e.equals(edge)) {
				return e.getLength();
			}
		}
		return -1;
	}
	
	/* * Normalizes so that the vector of all edges (both internal and ones ending in leaves) has length 1.
	 * 
	 */
	public void normalize() {
		double vecLength = 0;
		// first compute the current length of the vector of edge lengths
		for (int i =0; i < leaf2NumMap.size(); i++) {
			vecLength = vecLength + Math.pow(leafEdgeLengths[i],2);
		}
		vecLength = vecLength + Math.pow(getDistanceFromOrigin(),2);
		
		vecLength = Math.sqrt(vecLength);
		
		// divide by the length of the edge length vector to normalize
		for (int i =0; i < leaf2NumMap.size(); i++) {
			leafEdgeLengths[i] = leafEdgeLengths[i]/vecLength;
		}
		for (int i = 0; i < edges.size(); i++) {
			edges.get(i).setLength(edges.get(i).getLength()/vecLength);
		}
	}

	public void setLeaf2NumMap(Vector<String> leaf2NumMap) {
		this.leaf2NumMap = leaf2NumMap;
	}
	
	public double getDistanceFromOrigin() {
		double dist = 0;
		for (int i = 0; i< edges.size(); i++) {
			dist = dist + Math.pow(edges.get(i).getLength(), 2);
		}
		return Math.sqrt(dist);
	}
	
	
	/** Returns a vector containing edges with the same partition in t1 and t2.  If t1 and t2 do not have the same leaf2NumMap,
	 *  then returns null.   If t1 and t2 have no common edges, then returns a vector of size 0.
	 *  We set the length of the returned edges to be the absolute value of the difference between the length
	 *  of the edge in t1 and the length of the edge in t2.
	 *  Doesn't assume t1 and t2 have same # of edges.
	 * @param t1
	 * @param t2
	 * @return
	 */
	public static Vector<PhyloTreeEdge> getCommonEdges(PhyloTree t1, PhyloTree t2) {	
		Vector<PhyloTreeEdge> commonEdges = new Vector<PhyloTreeEdge>();
		
		// if the two trees do not have the same leaf2NumMap
		if (!(t1.getLeaf2NumMap().equals(t2.getLeaf2NumMap()))){
			System.out.println("Error: the two trees do not have the same leaves!");
			System.out.println("First tree's leaves are " + t1.getLeaf2NumMap() );
			System.out.println("Second tree's leaves are " + t2.getLeaf2NumMap() );
			System.exit(1);
		}
		
		for (int i = 0; i < t1.edges.size(); i++) {
			for (int j = 0; j < t2.edges.size(); j++) {
				
				// check that the edge is not 0 (has no leaves and can be thought of as a null edge; ignore such edges)
				if ( !t1.getEdge(i).isZero() && !t2.getEdge(j).isZero() && (t1.getEdge(i).equals(t2.getEdge(j))) ) {
					double distCommonEdge = Math.abs(t1.getEdge(i).getLength() - t2.getEdge(j).getLength());
					commonEdges.add(new PhyloTreeEdge(t1.getEdge(i).getEdge(), distCommonEdge));
//					System.out.println("Added new version of edge " + i + " from t1 to common edges; edge: " + t1.getEdge(i).getEdge() );
//					System.out.println("In for loop, commonEdges is " + commonEdges);
				}
			}
		}
//		System.out.println("returning these edges from getCommonEdges:" + commonEdges);
		return commonEdges;
	}
	
	public void permuteLeaves() {
		Vector<PhyloTreeEdge> v = this.edges;
		int numEdges = v.size();
		int numLeaves = numEdges + 2;  // because assuming trees are rooted, or have been rooted
		
//		System.out.println("numEdges: " + numEdges + "; v is " + v);
		
		Vector<PhyloTreeEdge> permutedV = new Vector<PhyloTreeEdge>();
		Random r = new Random();
		
		// set all the original edges and lengths
		for (int i =0; i < numEdges; i++) {
			permutedV.add(new PhyloTreeEdge(v.get(i).getLength() ) );
			permutedV.get(i).setOriginalEdge((BitSet) permutedV.get(i).getEdge().clone());
		}
		
		// initialize the vector to keep track of which leaves have already been permuted.
		Vector<Integer> leavesToBePermuted = new Vector<Integer>();
		for (int i = 0; i < numLeaves; i++ ) {
			leavesToBePermuted.add(new Integer(i));
		}
		
		// each time through the loop, one leaf is permuted. 
		for (int i = numLeaves; i > 0; i--) {
			int leaf = leavesToBePermuted.remove(r.nextInt(i)).intValue();
//			System.out.println("Moving leaf " + leaf + " to " + (i-1));

			// make the v.size() - i column in permutedV be col in v
			for (int j = 0; j < numEdges; j++ ) {
				// if the bit in the i-th column of the j-th row is 1 ...
				if (v.get(j).getEdge().get(leaf) ) {
					permutedV.get(j).getEdge().set(i-1);
				}
			}
		}
//		System.out.println("permutedV is " + permutedV);
		edges = TreeDistance.myVectorClonePhyloTreeEdge(permutedV);
	}
	
	
	/** Deletes the edge corresponding to bipartition e from the tree.
	 * 
	 * @param e
	 */
	public void removeEdge(Bipartition e){
		for(int i = 0; i < edges.size(); i++) {
			if (edges.get(i).sameBipartition(e)) {
				edges.remove(i);
				break;
			}
		}
	}
	
	/**  Removes the edges corresponding to the ones in v.
	 * 
	 * @param v
	 */
	public void removeEdgesIndicatedByOnes(Bipartition v) {
		int numRemovedSoFar = 0;
		for (int i = 0; i < leaf2NumMap.size() -2; i++) {
			if (v.contains(i)) {
				edges.remove(i - numRemovedSoFar);
				numRemovedSoFar++;
			}
		}
	}
	
	/** Adds the edge e to the tree
	 * 
	 * @param e
	 */
	public void addEdge(PhyloTreeEdge e) {
		edges.add(e);
	}
	
	/** Returns the smallest index in a string t after position i, at which one of the characters in s is located.
	 * 
	 * @param i
	 * @param s
	 * @return t.length() if no character in s is located after position i in t; otherwise return the min index
	 */
	public static int nextIndex(String t, int i, String s) {
		int minIndex = t.length();
		int tempIndex = -1;
		
		for (int j = 0; j < s.length(); j++) {
			tempIndex = t.substring(i+1).indexOf(s.charAt(j));
			if ((tempIndex != -1)  && (tempIndex + i + 1 < minIndex)) {
				minIndex = tempIndex + i +1;
			}
		}
		return minIndex;
	}
	
	
	/** Returns a vector containing representations of 0-1 vectors which
	 * have a 1 in coordinate i if edge i of tree t crosses the edge of this tree, which this 0-1 
	 * vector represents.  The entries of this vector correspond to the edges of this tree.
	 * Doesn't assume the trees have the same number of edges.
	 * XXX maybe should move exiting if trees don't have same leaf2numMap to this method??
	 * @param t
	 * @return
	 */
	public Vector<Bipartition> getCrossingsWith(PhyloTree t) {
//		System.out.println("this tree is " + this);
		// if the two trees do not have the same leaf2NumMap
		if (!(this.getLeaf2NumMap().equals(t.getLeaf2NumMap()))){
			return null;
		}
		
/*		Vector<Bipartition> edges = new Vector<Bipartition>();
		Iterator<PhyloTreeEdge> thisIter = this.edges.iterator();
		int i = 0;  // keeps track of which vector coordinate the edge 
		// that we are comparing with will correspond to
		
		while (thisIter.hasNext()) {
			PhyloTreeEdge thisTreeEdge = thisIter.next();
			Bipartition newEdge = new Bipartition();
			
			Iterator<PhyloTreeEdge> tIter = t.edges.iterator();
			i = 0;  // reset i
			while (tIter.hasNext()) {
				PhyloTreeEdge e = (PhyloTreeEdge) tIter.next();
				if (e.getEdge() != 0 && e.crosses(thisTreeEdge)) {
					newEdge.addOne(i);
				}
				i++;
			}
			edges.add(newEdge);
		}
		*/
//		System.out.println("in getCrossings, this tree is " + this);
		Vector<Bipartition> edges = new Vector<Bipartition>();
		// for each vector in m 
//		System.out.println("this.edges.size() is " + this.edges.size());
		for (int i = 0; i < this.edges.size(); i++) {
			
			PhyloTreeEdge thisTreeEdge = this.edges.get(i);;
			Bipartition newEdge = new Bipartition();
			
			// add the appropriate 1's
			for (int j = 0; j < t.edges.size(); j++) {
				PhyloTreeEdge e = t.edges.get(j);
				// e should not equal 0.
				if (e.crosses(thisTreeEdge)) {
					newEdge.addOne(j);
				}
			}
			edges.add(newEdge.clone());
//			System.out.println("Added edge " + newEdge.clone());
		}
//		System.out.println("in getCrossingsWith returning " + edges);
		return edges;
	}
	
	public String toString() {
		return "Leaves: " + leaf2NumMap + "; edges: " + edges; 
	}
	
	public PhyloTree clone() {
		return new PhyloTree(TreeDistance.myVectorClonePhyloTreeEdge(edges), TreeDistance.myVectorCloneString(leaf2NumMap), leafEdgeLengths.clone() );
	}
	
	// XXX for now this is the testing procedure
	public static void main(String args[]) {
		PhyloTree t1 = new PhyloTree("((A:0.1,B:0.2):1,(C:0.3,D:0.4):2)");
		// should produce:  [1.0 11, 2.0 1100]
		System.out.println("Edges of tree 1 (length, followed by leaves as 1's): " + t1.getEdges().toString());
		
		PhyloTree t2 = new PhyloTree("(((B:0.1,A:0.2):1,C:0.3):2,(D:0.4,E:0.5):3)");
		// should produce: [1.0 11, 2.0 111, 3.0 11000]
		System.out.println("Edges of tree 2: " + t2.getEdges().toString());
	
		// has the same leaf2NumMap as t2 and a common edge
		PhyloTree t3 = new PhyloTree("((A:0.1,B:0.2):1,(C:0.3,(D:0.4,E:0.5):2):3)");
		// should produce [1.0 11, 2.0 11100, 3.0 11000]
		System.out.println("Edges of tree 3: " + t3.getEdges().toString());
		
		System.out.println("common edges of t2 and t3: " + getCommonEdges(t2,t3));
		
		// should have no common edges with t2
		PhyloTree t4 = new PhyloTree("((A:0,C:0):1,(D:0,(B:0,E:0):2):3)");
		System.out.println("Edges of tree 4:" + t4.getEdges().toString());
		System.out.println("1s indicate edges in tree 4 crossed by that edge in tree2: " + t2.getCrossingsWith(t4));
		System.out.println("1s indicate edges in tree 1 crossed by that edge in tree 4: " + t1.getCrossingsWith(t4));
		
		PhyloTree triTop = new PhyloTree("(((A:0,B:0):1,(C:0,D:0):2):3,E:0)");
		System.out.println("Edges of tree TriTop: " +triTop.getEdges());
		PhyloTree triBot = new PhyloTree("(((D:0,E:0):1,(C:0,B:0):2):3,A:0)");
		System.out.println("1s indicate edgdes in tree triBot crossed by that edge in tree triTop:" + triBot.getCrossingsWith(triTop));
		PhyloTree noCommonSortingWithTriTop = new PhyloTree("(((A:0,C:0):1,E:0):2,(B:0,D:0):3)");
		System.out.println("Edges of tree noCommonSortingWithTriTop: " + noCommonSortingWithTriTop.getEdges());
		System.out.println("1s indicate edges in tree triTop crossed by that edge in tree noCommonSortingWithTriTop: " + 
				noCommonSortingWithTriTop.getCrossingsWith(triTop));
	}

	public double[] getLeafEdgeLengths() {
		return leafEdgeLengths;
	}

	public void setLeafEdgeLengths(double[] leafEdgeLengths) {
		this.leafEdgeLengths = leafEdgeLengths;
	}

	public String getNewick() {
		return newick;
	}

	public void setNewick(String newick) {
		this.newick = newick;
	}
}
