/** This file is part of GeodeMAPS, a program for computing the geodesic distance between phylogenetic trees.
    Copyright (C) 2008  Megan Owen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */



package distanceAlg1;

import java.util.*;

public class PhyloTreeEdge extends Bipartition {
	private double length;
	private Bipartition originalEdge;
	
	/** Constructor: creates a new edge with no child leaves and length 0
	 */
	public PhyloTreeEdge() {
		super();
		length = 0;
		originalEdge = new Bipartition();
	}
	
	/** Constructor: creates a new edge with the specificed child leaves and length 0
	 * 
	 * @param edge
	 */
	public PhyloTreeEdge(BitSet edge) {
		super(edge);
		length = 0;
		originalEdge = new Bipartition();
//		System.out.println("edge " + edge);
	}
	
	public PhyloTreeEdge(double length) {
		super();
		this.length = length;
		originalEdge = new Bipartition();
//		System.out.println("double length ");
	}
	
	public PhyloTreeEdge(double length, Bipartition originalEdge) {
		super();
		this.length = length;
		this.originalEdge = new Bipartition(originalEdge.edge);
	}
	
	public PhyloTreeEdge(BitSet edge, double length) {
		super(edge);
		this.length = length;
		originalEdge = new Bipartition(edge);
	}
	
	public PhyloTreeEdge(Bipartition edge, double length) {
		super(edge.edge);
		this.length = length;
		originalEdge = new Bipartition(edge.edge);
	}
	
	public PhyloTreeEdge(BitSet edge, double length, BitSet originalEdge) {
		super(edge);
		this.length = length;
		this.originalEdge = new Bipartition(originalEdge);
	}
	
	// Getters and Setters
	public double getLength() {
		return length;
	}

	public void setLength(double length) {
		this.length = length;
	}
	
	/** String representation of an edge.  
	 * Returns the length followed by a 0-1 vector, representing the children.
	 */
/*	public String toString() {
		return "" + TreeDistance.truncate(length,6) + " " + super.toString();
	}
*/	
	/** String representation of an edge.  
	 * Returns the length as the original integer representing all children of this edge.
	 * If the original edge hasn't been filled in (is 0), then return the actually edge.
	 */
	public String toString() {
		if (originalEdge.isZero()) {
			return ""+ TreeDistance.truncate(length,3) + " " + edge;
		}
		return "" + TreeDistance.truncate(length,3) + " " + edge;
//		return "" + TreeDistance.truncate(length,3) + " " + TreeDistance.binaryToVector(originalEdge.getEdge());
	}
	
	public String toStringVerbose(Vector<String> leaf2NumMap) {
		return "" + TreeDistance.truncate(length, 6) + "\t\t" + Bipartition.toStringVerbose(this.edge, leaf2NumMap);
	}
	
	/** Returns a string of what it printed out.
	 * 
	 * @param edges
	 * @param leaf2NumMap
	 * @return
	 */
	public static String printEdgesVerbose(Vector<PhyloTreeEdge> edges, Vector<String> leaf2NumMap, Boolean originalEdges) {
		String output = "";
		
		System.out.println("Length\t\tEdge");
		output = output + "Length\t\tEdge\n";
		for (int i =0; i < edges.size();i++) {
			if (edges.get(i).getOriginalEdge().getEdge() != null && originalEdges) {
				System.out.println(TreeDistance.truncate(edges.get(i).length, 6) + "\t\t" +  Bipartition.toStringVerbose(edges.get(i).getOriginalEdge().getEdge(), leaf2NumMap));
				output = output + TreeDistance.truncate(edges.get(i).length, 6) + "\t\t" +  Bipartition.toStringVerbose(edges.get(i).getOriginalEdge().getEdge(), leaf2NumMap) + "\n";
			}
			else {
				System.out.println(edges.get(i).toStringVerbose(leaf2NumMap));
				output = output + edges.get(i).toStringVerbose(leaf2NumMap) + "\n";
			}
		}
		return output;
	}
	
	
	public PhyloTreeEdge clone() {
		return new PhyloTreeEdge(this.edge, this.length, this.originalEdge.getEdge());
	}
	
	public boolean sameBipartition(PhyloTreeEdge e) {
		return this.edge == e.edge;
	}
	
	public boolean sameBipartition(Bipartition e) {
		return this.edge == e.edge;
	}

	public Bipartition getOriginalEdge() {
		return originalEdge;
	}

	public void setOriginalEdge(Bipartition originalEdge) {
		this.originalEdge = originalEdge;
	}
	
	public void setOriginalEdge(BitSet originalEdge) {
		this.originalEdge = new Bipartition(originalEdge);
	}
}
